/*
 * Copyright (c) 2001, 2003 IBM Corp and others.  All rights reserved.
 * This file is made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 * 
 * Contributors: 
 * 	  Andre Weinand (OTI Labs)
 */
 
/* MacOS X Carbon specific logic for displaying the splash screen. */

#include "eclipseOS.h"

#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <ctype.h>
#include <pwd.h>

#include <CoreServices/CoreServices.h>
#include <Carbon/Carbon.h>

#include "NgCommon.h"
#include "NgImageData.h"

#define JAVA_SWT "/java_swt"
#define startupJarName "startup.jar"
#define MAX_PATH_LENGTH 2000
#define DEBUG 0

char *findCommand(char *command);

static char *strsave(char *s);
static void debug(const char *fmt, ...);
static void dumpArgs(char *tag, int argc, char* argv[]);
static PixMapHandle loadBMPImage(const char *image);
static char* getInstallDir2(const char *program);
static char* expandTilde(char *arg);


/* Global Variables */
char   dirSeparator  = '/';
char   pathSeparator = ':';
char*  consoleVM     = "java";
char*  defaultVM     = "java";
char*  shippedVMDir  = "jre/bin/";
char*  guiList[]	 = { "carbon" };

/* Define the window system arguments for the various Java VMs. */
static char*  argVM_JAVA[] = { NULL };

#define MAGIC_FILE "/tmp/eclipse_magic_%s"	/* psn is appended*/
#define MAGIC "org.eclipse.core.boot"
static int fgPid;
static FILE *fgConsoleLog;
static char *fgPsn;
static char *fgProgram;
static int fgRestarted;


int main( int argc, char* argv[] ) {

	int i;
	
	fgConsoleLog= fopen("/dev/console", "w");
	fgProgram= strsave(argv[0]);
	fgPid= getpid();
	
	dumpArgs("start", argc, argv);
		
	char *k= "-showsplash_";
	int l= strlen(k);
	for (i= 0; i < argc; i++) {
		char *arg= argv[i];
		if (strncmp(arg, k, l) == 0) {
			fgPsn= &arg[l];
			argv[i]= "-showsplash";
		}
	}	
		
	if (argc > 1 && strncmp(argv[1], "-psn_", 5) == 0) {
	
		fgPsn= &argv[1][5];
	
		// Get the main bundle for the app
		CFBundleRef mainBundle= CFBundleGetMainBundle();
		
		if (mainBundle != NULL) {
			
			// Get an instance of the info plist.
			CFDictionaryRef bundleInfoDict= CFBundleGetInfoDictionary(mainBundle);
			
			// If we succeeded, look for our property.
			if (bundleInfoDict != NULL) {
			    CFArrayRef ar= CFDictionaryGetValue(bundleInfoDict, CFSTR("Eclipse"));
			    CFIndex size= CFArrayGetCount(ar);
			    if (size > 0) {
			    	int i;
					char **old_argv= argv;
					argv= (char**) calloc(size+2, sizeof(char*));
					argc= 0;
					
					argv[argc++]= old_argv[0];	// $0
					argv[argc++]= old_argv[1];	// -psn_major_minor
					
				    for (i= 0; i < size; i++) {
						CFStringRef sr= (CFStringRef) CFArrayGetValueAtIndex (ar, i);
            			CFIndex argStringSize= CFStringGetMaximumSizeForEncoding(CFStringGetLength(sr), kCFStringEncodingUTF8);
						char *s= (char*) malloc(argStringSize);
           				if (CFStringGetCString(sr, s, argStringSize, kCFStringEncodingUTF8)) {
           					argv[argc++]= expandTilde(s);				
            			} else {
							fprintf(fgConsoleLog, "can't extract bytes\n");
            			}
           				free(s);
				    }
			    }
			} else {
				fprintf(fgConsoleLog, "no bundle dict found\n");
			}
		} else {
			fprintf(fgConsoleLog, "no bundle found\n");
		}
	}
	
	// if a Java 1.4.1 VM is specified, replace it with our swt_java launcher
	for (i= 0; i < argc; i++) {
		char *arg= argv[i];
		if (strcmp(arg, "-vm") == 0) {
			i++;
			if (i < argc && strstr(argv[i], "1.4.1") != NULL) {
				
				char *program= findCommand(argv[0]);
				char *buf= calloc(1, strlen(program) + strlen(JAVA_SWT) + 4);
				strcpy(buf, program);
				
				char *p= rindex(buf, '/');
				if (p != NULL) {
					*p= '\0';
					strcat(buf, JAVA_SWT);
					// fprintf(stderr, "VM: %s\n", buf);
					argv[i]= buf;
					// make sure, its executable
					chmod(buf, 0755);
				}
			}
			break;
		}
	}
	
	int exitcode= original_main(argc, argv);
	
	if (fgPsn != NULL) {
		char magicFilename[200];
		sprintf(magicFilename, MAGIC_FILE, fgPsn);
		debug("removing magic file: <%s>\n", magicFilename);
		unlink(magicFilename);
	}
	
	debug("<<<< exit(%d)\n", exitcode);
	
	fclose(fgConsoleLog);

	return exitcode;
}

/*
	A word beginning with an unquoted tilde character (~) is
	subjected to tilde expansion.  All the characters up to a
	slash (/) or the end of the word are treated as a username
	and are replaced with the user's home directory.  If the
	username is missing (as in ~/foobar), the tilde is
	replaced with the value of the HOME variable (the current
	user's home directory).
*/
static char* expandTilde(char *arg) {
	char buffer[1000];
	char c, lastChar= ' ', *cp= arg;
	
	//debug("in: <%s>\n", arg);
	int out= 0;
	while (c= *cp++) {
		if (isspace(lastChar) && c == '~') {
			char name[100], *dir= NULL;
			int j= 0;
			for (; c= *cp; cp++) {
				if (! isalnum(c))
					break;
				name[j++]= c;
				lastChar= c;
			}
			name[j]= '\0';
			if (j > 0) {
				struct passwd *pw= getpwnam(name);
				if (pw != NULL)
					dir= pw->pw_dir;
			} else {
				dir= getenv("HOME");
			}
			if (dir != NULL) {
				strcpy(&buffer[out], dir);
				out+= strlen(dir);
			}
		} else 
			buffer[out++]= c;
		lastChar= c;
	}
	buffer[out]= '\0';
	//debug("  out: <%s>\n", buffer);
	return strsave(buffer);
}

/* Display a Message */
void displayMessage( char* message )
{
	SInt16 outItemHit;
	int l= strlen(message);
	if (l > 255)
		l= 255;
	char *buffer= malloc(l + 1);
	buffer[0]= l;
	strncpy(&buffer[1], message, l);
	
	if (StandardAlert(kAlertStopAlert, buffer, NULL, NULL, &outItemHit) != 0)
		debug("eclipse: displayMessage: %s\n", message);
		
	free(buffer);
}

static char *strsave(char *s) {
	return strcpy(malloc(strlen(s)+1), s);
}

static void debug(const char *fmt, ...) {
#if DEBUG
	char buffer[200];
	va_list ap;
	va_start(ap, fmt);
	fprintf(fgConsoleLog, "%05d: ", fgPid);
	vfprintf(fgConsoleLog, fmt, ap);
	va_end(ap);
#endif
}

static void dumpArgs(char *tag, int argc, char* argv[]) {
#if DEBUG
	int i;
	if (argc < 0) {
		argc= 0;
		for (i= 0; argv[i] != NULL; i++)
			 argc++;
	}
	debug(">>>> %s:", tag);
	for (i= 0; i < argc && argv[i] != NULL; i++)
		fprintf(fgConsoleLog, " %s", argv[i]);
	fprintf(fgConsoleLog, "\n");
#endif
}

static void init() {
	static int initialized= 0;
	
	if (!initialized) {
		ProcessSerialNumber psn;
		if (GetCurrentProcess(&psn) == noErr) {
			//CPSEnableForegroundOperation(&psn, 0x03, 0x3C, 0x2C, 0x1103);
			SetFrontProcess(&psn);
		}
		ClearMenuBar();
		initialized= true;
	}
}

/* Initialize Window System
 *
 * Initialize Carbon.
 */
void initWindowSystem( int* pArgc, char* argv[], int showSplash )
{
    char *program= findCommand(argv[0]);
    debug("program: %s\n", program);
    char *homeDir= getInstallDir2(program);
    if (homeDir != NULL)
    	chdir(homeDir);
    
	if (showSplash)
		init();
}

static void eventLoopTimerProc(EventLoopTimerRef inTimer, void *inUserData) {
	QuitApplicationEventLoop();
}

/* Show the Splash Window
 *
 * Create the splash window, load the bitmap and display the splash window.
 */
int showSplash( char* timeoutString, char* homeDir, char* featureImage )
{
    Rect wRect, deviceRect;
    WindowRef myWindow= NULL;
    EventRecord	event;
	int w, h, deviceWidth, deviceHeight;
	PixMap *pm;
	PixMapHandle pixmap;
	OSStatus status;
	
	/*
	 * Calling chain:
	 * eclipse.exe <Eclipse>
	 *      /usr/bin/java
	 *             eclipse.exe <showSplash>
	 *      /usr/bin/java
	 *             eclipse.exe <showSplash>
	 *
	 * In this the second "eclipse.exe showSplash" must be launched differently than the first
	   we have to detect the case where 
	// HACK: we are setting a global flag here by creating a temporary file
	 */
	debug("featureImage: %s\n", featureImage);
	if (featureImage != NULL && strstr(featureImage, MAGIC) != NULL && fgPsn != NULL) {	
		char magicFilename[200];
		sprintf(magicFilename, MAGIC_FILE, fgPsn);
		debug("creating magic file: <%s>\n", magicFilename);
		if (open(magicFilename, O_WRONLY | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR) == -1)
			debug("error creating magic file (%s): %d\n", magicFilename, errno);
	}
	
	init();
    
	/* Determine the splash timeout value (in seconds). */
	if (timeoutString != NULL && strlen(timeoutString) > 0) {
    	int timeout;
	    if (sscanf(timeoutString, "%d", &timeout) == 1) {
	    	EventLoopTimerUPP upp= NewEventLoopTimerUPP(eventLoopTimerProc);
			InstallEventLoopTimer(GetMainEventLoop(), (EventTimerInterval) timeout, 0.0, upp, NULL, NULL);
		}
	}
	
	pixmap= loadBMPImage(featureImage);
    /* If the splash image data could not be loaded, return an error. */
	if (pixmap == NULL)
		return ENOENT;
		
	pm= *pixmap;
	w= pm->bounds.right;
	h= pm->bounds.bottom;
	
	GetAvailableWindowPositioningBounds(GetMainDevice(), &wRect);
	
	deviceWidth= wRect.right - wRect.left;
	deviceHeight= wRect.bottom - wRect.top;
	
	wRect.left+= (deviceWidth-w)/2;
	wRect.top+= (deviceHeight-h)/3;
	wRect.right= wRect.left + w;
	wRect.bottom= wRect.top + h;
    
	CreateNewWindow(kModalWindowClass, kWindowNoAttributes, &wRect, &myWindow);	
    if (myWindow != nil) {
		GrafPtr port= GetWindowPort(myWindow);
		SetPort(port);  /* set port to new window */
     	SetRect(&wRect, 0, 0, w, h);
		ShowWindow(myWindow);
		CopyBits((BitMap*)pm, GetPortBitMapForCopyBits(port), &wRect, &wRect, srcCopy, NULL);

	    RunApplicationEventLoop();
	    
	    DisposeWindow(myWindow);
	}
    
    return 0;
}


/* Get the window system specific VM arguments */
char** getArgVM( char* vm ) 
{
    char** result;
    
    /* Use the default arguments for a standard Java VM */
    result = argVM_JAVA;
    
    return result;
}

/* Start the Java VM 
 *
 * This method is called to start the Java virtual machine and to wait until it
 * terminates. The function returns the exit code from the JVM.
 */
int startJavaVM( char* args[] ) 
{

	if (fgRestarted > 0) {
		// test whether the magic file has been created;
		char magicFilename[200];
		sprintf(magicFilename, MAGIC_FILE, fgPsn);
		debug("testing existence of file: <%s>\n", magicFilename);
		struct stat sb;
		if (stat(magicFilename, &sb) == -1) {
			debug("  file <%s> does not exist\n", magicFilename);
			char *p= strstr(fgProgram, "/Contents/MacOS");
			if (p != NULL) {
				char buf[400];
				*p= '\0';
				sprintf(buf, "/usr/bin/open \"%s\"", fgProgram);
				debug("system(%s)\n", buf);
				system(buf);
				debug("<<<< exit(0)\n");
				exit(0);
			}
		} else {
			unlink(magicFilename);
			debug("  file <%s> does exist\n", magicFilename);
		}
	}

	/* Create a child process for the JVM. */	
	pid_t pid= fork();
	if (pid == 0) {
		int i;
	
		for (i= 0; args[i] != NULL; i++) {
			char *arg= args[i];
			if (fgPsn != NULL && strcmp(arg, "-showsplash") == 0) {
				char *arg= args[i+1];
				char buf1[200];
				char buf2[200];
				char buf3[200];
				if (sscanf(arg, "%s %s %s", buf1, buf2, buf3) == 3) {
					if (strcmp(buf2, "-showsplash") == 0) {
						char buf[1000];
						sprintf(buf, "%s -showsplash_%s %s", buf1, fgPsn, buf3);
						args[i+1]= strsave(buf);
					}
				}
			}
		}
	
		dumpArgs("execv", -1, args);

		/* Child process ... start the JVM */
		execv(args[0], args);
		
		/* The JVM would not start ... return error code to parent process. */
        _exit(errno);
	}
	
	if (pid == -1)
		return errno;
	
	/* wait for it to terminate. */
    int exitCode;
	wait(&exitCode);
		
	exitCode= ((exitCode & 0x00ff) == 0 ? (exitCode >> 8) : exitCode);
	
	if (exitCode == 23)	
		fgRestarted= 1;
	
	return exitCode;
}

/////////////////////////////////////////

/**
 * loadBMPImage
 * Create a QuickDraw PixMap representing the given BMP file.
 *
 * bmpPathname: absolute path and name to the bmp file
 *
 * returned value: the PixMapHandle newly created if successful. 0 otherwise.
 */
static PixMapHandle loadBMPImage (const char *bmpPathname) { 
	ng_stream_t in;
	ng_bitmap_image_t image;
	ng_err_t err= ERR_OK;
	PixMapHandle pixmap;
	PixMap *pm;

	NgInit();
		
	if (NgStreamInit(&in, (char*) bmpPathname) != ERR_OK) {
		NgError(ERR_NG, "Error can't open BMP file");
		return 0;
	}
	
	NgBitmapImageInit(&image);
	err= NgBmpDecoderReadImage (&in, &image);
	NgStreamClose(&in);
	
	if (err != ERR_OK) {
		NgBitmapImageFree(&image);
		return 0;
	}
	
	UBYTE4 srcDepth= NgBitmapImageBitCount(&image);
	if (srcDepth != 24) {	/* We only support image depth of 24 bits */
		NgBitmapImageFree(&image);
		NgError (ERR_NG, "Error unsupported depth - only support 24 bit");
		return 0;
	}
	
	pixmap= NewPixMap();	
	if (pixmap == 0) {
		NgBitmapImageFree(&image);
		NgError(ERR_NG, "Error XCreatePixmap failed");
		return 0;
	}
	
	pm= *pixmap;
	
	int width= (int)NgBitmapImageWidth(&image);
	int height= (int)NgBitmapImageHeight(&image);
	int rowBytes= width * 4;
	
	pm->bounds.right= width;
	pm->bounds.bottom= height;
	pm->rowBytes= rowBytes + 0x8000; 
	pm->baseAddr= NewPtr(rowBytes * height);
	pm->pixelType= RGBDirect;
	pm->pixelSize= 32;
	pm->cmpCount= 3;
	pm->cmpSize= 8;
	
	/* 24 bit source to direct screen destination */
	NgBitmapImageBlitDirectToDirect(NgBitmapImageImageData(&image), NgBitmapImageBytesPerRow(&image), width, height,
		(UBYTE1*)pm->baseAddr, pm->pixelSize, rowBytes, NgIsMSB(),
			0xff0000, 0x00ff00, 0x0000ff);
			
	NgBitmapImageFree(&image);
	
	return pixmap;
}

/* Determine the Installation Directory
 *
 * This function takes the directory where program executable resides and
 * determines the installation directory. The installation directory must
 * contain the startup JAR file and the executable must be either in the 
 * install directory or in a subdirectory.
 */
static char* getInstallDir2(const char *program) 
{
	char*  ch;
	char*  installDir;
	char   path[ MAX_PATH_LENGTH ];
    struct stat stats;
	
	/* Start with the directory where the executable resides. */
    installDir = (char*) malloc( strlen( program ) + 1 );
    strcpy( installDir, program );
    ch = strrchr( installDir, (int)dirSeparator );
    
    /* For each directory in the hierarchy */
    while (ch != NULL) {
    	/* Format the pathname of the startup jar file. */
    	*(ch+1) = '\0';
    	sprintf( path, "%s%s", installDir, startupJarName );
    	
    	/* If the jar file exists */
    	if (stat( path, &stats ) == 0 && (stats.st_mode & S_IFREG) != 0) {
    		return installDir;
    	} else {
    		/* Try the next directory up in the hierarchy. */
    		*ch = '\0';		
    		ch = strrchr( installDir, (int)dirSeparator );
    	}
    }
    
    free( installDir );
    return NULL;
}

